#include "gcl/Geometry.h"
#include "gcl/GLXPane.h"
#include "gcl/VecUtil.h"
#include "gcl/Draw.h"

#ifdef TEST
#define DBG_COUT if (1) cout
Void ShowPoly(CoordList &poly);
Void ShowPolyRect(CoordList &poly1, CoordList &poly2, Int *id = 0);
#else
#define DBG_COUT if (0) cout
#endif

Void FindConvexHull(CoordList &poly, Int minIdx[2], Int maxIdx[2])
// find convex hull of poly, and indices of the min & max points
// in x and y. O(n).
// XXX speed up by doing in place &etc.
{
	static CoordList stack;
	Coord	e1, e2;
	Coord	min, max, last, curr, next;
	GCLReal	cp, xMin;
	Int		i;
			
	min = poly[0];
	minIdx[0] = 0;

	// find left-most point; this is definitely on the convex hull
	for (i = 1; i < poly.NumItems(); i++)
		if (poly[i][0] < min[0])
		{
			min[0] = poly[i][0];
			minIdx[0] = i;
		}		
		
	// use that as our starting point
	stack.SetSize(0);
	stack.Push(poly[minIdx[0]]);
	i = (minIdx[0] + 1) % poly.NumItems();

	// now, scan around the outside
	while (i != minIdx[0])
	{
		// push vertex on stack
		stack.Push(poly[i]);
		i = (i + 1) % poly.NumItems();

		// using next vertex, keep popping off vertices on the
		// stack until the last two vertices form a convex
		// angle with it.
		do 
		{
			e1 = stack.Top() - stack[stack.NumItems() - 2];
			e2 = poly[i] - stack.Top();
			cp = e1[0] * e2[1] - e1[1] * e2[0];

			if (cp < 0)
				stack.Pop();
		} while (cp < 0 && stack.NumItems() > 1);
	}
			

	poly = stack;

	minIdx[0] = minIdx[1] = maxIdx[0] = maxIdx[1] = 0;
	min = max = poly[0];

	for (i = 1; i < poly.NumItems(); i++)
	{
		if (poly[i][0] > max[0])
		{
			max[0] = poly[i][0];
			maxIdx[0] = i;
		}
		if (poly[i][1] < min[1])
		{
			min[1] = poly[i][1];
			minIdx[1] = i;
		}
		else if (poly[i][1] > max[1])
		{
			max[1] = poly[i][1];
			maxIdx[1] = i;
		}
	}
}

Void FindExtrema(CoordList &coords, Int minIdx[2], Int maxIdx[2])
{
	Coord	min, max;
	Int		i;

	minIdx[0] = minIdx[1] = maxIdx[0] = maxIdx[1] = 0;
	min = max = coords[0];

	for (i = 1; i < coords.NumItems(); i++)
	{
		if (coords[i][0] < min[0])
		{
			min[0] = coords[i][0];
			minIdx[0] = i;
		}
		else if (coords[i][0] > max[0])
		{
			max[0] = coords[i][0];
			maxIdx[0] = i;
		}
		if (coords[i][1] < min[1])
		{
			min[1] = coords[i][1];
			minIdx[1] = i;
		}
		else if (coords[i][1] > max[1])
		{
			max[1] = coords[i][1];
			maxIdx[1] = i;
		}
	}
}

#ifdef UNFINISHED
Void FindConvexHullUnordered(CoordList &coords)
// find convex hull of unordered points, and indices of the min & max points
// in x and y. O(n).
{
	static CoordList stack;
	Coord	e1, e2;
	Coord	min, max, last, curr, next;
	GCLReal	cp, xMin;
	Int		i;
			
	// use that as our starting point
	stack.SetSize(0);
	stack.Push(coords[minIdx[0]]);
	i = (minIdx[0] + 1) % coords.NumItems();

	// now, scan around the outside
	while (i != minIdx[0])
	{
		// push vertex on stack
		stack.Push(coords[i]);
		i = (i + 1) % coords.NumItems();

		// using next vertex, keep popping off vertices on the
		// stack until the last two vertices form a convex
		// angle with it.
		do 
		{
			e1 = stack.Top() - stack[stack.NumItems() - 2];
			e2 = coords[i] - stack.Top();
			cp = e1[0] * e2[1] - e1[1] * e2[0];

			if (cp < 0)
				stack.Pop();
		} while (cp < 0 && stack.NumItems() > 1);
	}
			
	coords = stack;
}
#endif

inline Int MaxEltIndex(const Vec4d &v)
{
	if (v[0] > v[1])
		if (v[0] > v[2])
			if (v[0] > v[3])
				return(0);
			else
				return(3);
		else
			if (v[2] > v[3])
				return(2);
			else
				return(3);
	else
		if (v[1] > v[2])
			if (v[1] > v[3])
				return(1);
			else
				return(3);
		else
			if (v[2] > v[3])
				return(2);
			else
				return(3);
}

/*

             e[3]
		+<-----------+
		|            ^
		|    v1      |
		|    ^       |
   e[0] |    |       | e[1]
		|    +-> v0  |
		|            |
		v            |
		+----------->+
             e[2]

*/
Void MinEnclRect(CoordList &hull, Int minIdx[2], Int maxIdx[2], Coord &minAxis)
// rotating callipers algorithm.
{
	Coord	e[4], v0, v1;
	Int		id[4], i;
	Vec4d	a;
	Double	minArea, l, area, height, width, scale;
	
	minAxis = Coord(-9.0, -9.0);
	
	v0 = vl_x;
	v1 = vl_y;
	
	id[0] = minIdx[0];
	id[1] = maxIdx[0];
	id[2] = minIdx[1];
	id[3] = maxIdx[1];
	e[0] = hull[(id[0] + 1) % hull.NumItems()] - hull[id[0]];
	e[1] = hull[(id[1] + 1) % hull.NumItems()] - hull[id[1]];
	e[2] = hull[(id[2] + 1) % hull.NumItems()] - hull[id[2]];
	e[3] = hull[(id[3] + 1) % hull.NumItems()] - hull[id[3]];
	
	scale = Max(hull[id[1]][0] - hull[id[0]][0], hull[id[3]][1] - hull[id[2]][1]);
	minArea = vl_inf;

	while (v0[0] > 0 && v1[1] > 0)
	{
		DBG_COUT << "this iter: " << id[0] << ' ' << id[1] << ' ' << id[2] << ' ' << id[3] << endl;
		
		for (i = 0; i < 4; i++)
		{
			l = len(e[i]);
			if (l/scale > 0.0001)
				e[i] /= l;
			else
				break;
		}

		if (i == 4)
		{
			// find next edge
			a[0] = -dot(e[0], v1);
			a[1] = dot(e[1], v1);
			a[2] = dot(e[2], v0);
			a[3] = -dot(e[3], v0);

			i = MaxEltIndex(a);

			DBG_COUT << "increments: " << a << endl;
			DBG_COUT << "picking " << i << endl;
		
			if (i == 0)
			{
				v1 = -e[0];
				v0 = cross(v1);
			}
			else if (i == 1)
			{
				v1 = e[1];
				v0 = cross(v1);
			}
			else if (i == 2)
			{
				v0 = e[2];
				v1 = -cross(v0);
			}
			else
			{
				v0 = -e[3];
				v1 = -cross(v0);
			}
		
			DBG_COUT << "new alignment: " << v0 << v1 << endl;
			width = dot(hull[id[1]] - hull[id[0]], v0);
			height = dot(hull[id[3]] - hull[id[2]], v1);
			area = width * height;
			DBG_COUT << "area = " << area << endl;

#ifdef TEST
			{ // --------------------------
				static CoordList rect(4);
				Coord rmin, rmax;
				
				rmin[0] = dot(v0, hull[id[0]]);
				rmax[0] = dot(v0, hull[id[1]]);
				rmin[1] = dot(v1, hull[id[2]]);
				rmax[1] = dot(v1, hull[id[3]]);
				rect[0] = v0 * rmin[0] + v1 * rmin[1];
				rect[1] = v0 * rmax[0] + v1 * rmin[1];
				rect[2] = v0 * rmax[0] + v1 * rmax[1];
				rect[3] = v0 * rmin[0] + v1 * rmax[1];

				ShowPolyRect(hull, rect, id);
			} // --------------------------
#endif
			
			if (area < minArea)
			{
				minArea = area;
				minAxis = v0;
			}
		}
		else
			DBG_COUT << "*** advancing along zero-length edge" << endl;
				
		DBG_COUT << "advanced from " << id[i] << " to " << (id[i] + 1) % hull.NumItems() << endl;
		id[i] = (id[i] + 1) % hull.NumItems();
		e[i] = hull[(id[i] + 1) % hull.NumItems()] - hull[id[i]];

		if (id[0] == id[1] || id[1] == id[2] || id[2] == id[3] || id[3] == id[0])
			DBG_COUT << "WARNING: overlap vtx" << endl;
	}

#ifdef TEST
	{ // --------------------------
		static CoordList rect(4);
		Coord rmin, rmax, v;
		CoordTrans m;
		Int	i;
		
		m[0] = minAxis;
		m[1] = -cross(minAxis);
		rmin.MakeBlock(vl_inf);
		rmax.MakeBlock(-vl_inf);

		for (i = 0; i < hull.NumItems(); i++)
		{
			v = m * hull[i];
			rmin[0] = Min(rmin[0], v[0]);
			rmax[0] = Max(rmax[0], v[0]);
			rmin[1] = Min(rmin[1], v[1]);
			rmax[1] = Max(rmax[1], v[1]);
		}

		m = trans(m);
		rect[0] = m * rmin;
		rect[1] = m * Coord(rmax[0], rmin[1]);
		rect[2] = m * rmax;
		rect[3] = m * Coord(rmin[0], rmax[1]);

		ShowPolyRect(hull, rect);
	} // --------------------------
#endif
}

#ifdef TEST
GLXOffscreenPane *gDrawPane;
XBackedPane *gTestWin;

Void DrawPoly(CoordList &poly, Renderer &r)
{
	Int		i;

	r.Begin(renLineLoop);
	for (i = 0; i < poly.NumItems(); i++)
		r.P(poly[i]);
	r.End();
}

Void ShowPoly(CoordList &poly)
{
	gDrawPane->Clear();
	DrawPoly(poly, *gDrawPane);
	gDrawPane->Show();
	gTestWin->HandleExpose();
	gTestWin->xgs->Run();
}

Void ShowPolyRect(CoordList &poly1, CoordList &poly2, Int *id = 0)
{
	gDrawPane->Clear().SetColour(cGreen);
	DrawPoly(poly1, *gDrawPane);
	gDrawPane->SetColour(cRed);
	DrawPoly(poly2, *gDrawPane);
	if (id)
	{
		Coord dv(0.05, 0.05);
		
		gDrawPane->SetColour(cBlue);
		FrameRect(*gDrawPane, poly1[id[0]] - dv, poly1[id[0]] + dv);
		FrameRect(*gDrawPane, poly1[id[1]] - dv, poly1[id[1]] + dv);
		FrameRect(*gDrawPane, poly1[id[2]] - dv, poly1[id[2]] + dv);
		FrameRect(*gDrawPane, poly1[id[3]] - dv, poly1[id[3]] + dv);
	}
	
	gDrawPane->Show();
	gTestWin->HandleExpose();
	gTestWin->xgs->Run();
}

class TestWindow : public XBackedPane
{
public:
	Void TrackMouse(Int x, Int y, Int modifiers)
	{
		if (!getClicks)
		{
			xgs->SignalDone();
			return;
		}

		Renderer *r = (GLXOffscreenPane*) backPane;
		clicks.Append(2 * Coord(x, height - 1 - y) / Coord(width, height) - vl_1);
		r->Clear();
		DrawPoly(clicks, *r);
		r->Show();
		HandleExpose();
	}

	Void HandleKey(Char c)
	{ xgs->SignalDone(); }
	
	Bool		getClicks;
	CoordList	clicks;
};

int main(int argc, char *argv[])
{
	XGraphicsSystem 	*xgs = new XGraphicsSystem;
	GLXOffscreenPane	*drawPane = new GLXOffscreenPane();
	TestWindow			*testWin = new TestWindow;
	CoordList 			poly, hull;
	Int					size = 200;
	Coord				minAxis;
	Int					minIdx[2], maxIdx[2];
	
	xgs->CreateOffscreen(drawPane, size, size);
	xgs->CreateWindow(testWin, "Geom2d test", size, size);
	testWin->Attach(drawPane);

	gDrawPane = drawPane;
	gTestWin = testWin;
	
	drawPane->SetBgColour(cWhite);
	drawPane->SetColour(cGreen);
	drawPane->Clear().Show();

	if (argc == 2)
		poly.ReadFile(argv[1]);
	else
	{
		testWin->getClicks = true;
		xgs->Run();
		poly = testWin->clicks;
		cout << poly << endl;	
		poly.WriteFile("last.poly");
	}
	
	testWin->getClicks = false;

	minIdx[0] = minIdx[1] = -1;
	maxIdx[0] = maxIdx[1] = -1;
	
	hull = poly;
	cout << "poly = " << hull << endl;
	FindConvexHull(hull, minIdx, maxIdx);
	cout << hull << endl;
	printf("%d %d %d %d\n", minIdx[0], maxIdx[0], minIdx[1], maxIdx[1]);

	drawPane->Clear();
	drawPane->SetColour(cRed);
	DrawPoly(hull, *drawPane);
	drawPane->SetColour(cBlack);
	DrawPoly(poly, *drawPane);
	drawPane->Show();
	testWin->HandleExpose();
	xgs->Run();

	MinEnclRect(hull, minIdx, maxIdx, minAxis);
		
	return(0);
}

#endif
