/*
	File:			SLContext.cc

	Function:		See header file

	Author(s):		Andrew Willmott

	Copyright:		Copyright (c) 1995-1996, Andrew Willmott

	Notes:			

*/


#include "gcl/SLContext.h"

/*!	\class SLContext

	State machine for rendering a scene; you can add attributes to the
	context, and also push & pop the current context.
	
	The context maintains a list of the attributes (i.e., state
	variables), and can conceptually pop or push that list from/to a
	stack. In fact attributes are only pushed / popped on a need-to
	basis, to avoid the overhead of copying the entire list each time.
*/


const UInt16 slc_EOL        = 0x8000;
const UInt16 slc_DaemonFlag = 0x4000;
const UInt16 slc_LevelMask  = 0x0FFF;


SLContext::SLContext(Int numAttrs) :
	attributeList(numAttrs),
	attributeStack()
{
	Clear();
}

AttrPtr SLContext::Get(AttrType attrType)
{
	if (attributeList[attrType].level & slc_DaemonFlag)
		return(attributeList[attrType].daemon->Get());
	else
		return(attributeList[attrType].data);		
}

Void SLContext::Set(AttrType attrType, AttrPtr attrPtr)
{
	Assert(attrType < attributeList.NumItems(),
			"(SLContext::Set) Unknown attribute type");
	AttrRec	*pRec = &attributeList[attrType];
	AttrStackRec	sRec;
	Bool			isDaemon = (pRec->level & slc_DaemonFlag);
				
	// if the attribute exists and wasn't set this level, push it...

	if (pRec->data && (pRec->level & slc_LevelMask) < level)
	{
		sRec.attrRec = *pRec;
		sRec.attrType = attrType;
		
		if (isDaemon)
			pRec->daemon->Save();

		attributeStack.Push(sRec);
	}
	
	if (isDaemon)
		pRec->daemon->Set(attrPtr);
	else
		pRec->data = attrPtr;

	pRec->level = level;
}

Void SLContext::SetDaemon(AttrType attrType, AttrDaemon *daemon)
{
	Assert((attributeList[attrType].level & slc_DaemonFlag) == 0, "Daemon already set for this slot.");
	
	Set(attrType, daemon);

	if (daemon)
		attributeList[attrType].level |= slc_DaemonFlag;
}

Void SLContext::Push()
{
	AttrStackRec sRec;
	
	level++;
	sRec.attrRec.level = slc_EOL;	// mark pRec as a sentinel
	attributeStack.Push(sRec);		// Add to the stack
}

Void SLContext::Pop()
{
	Int i;
	
//	pop off all the attributes changed since the last push...

	Assert(level > 0, "(SLContext::Pop) Too many pops");
	
	while (!(attributeStack.Top().attrRec.level & slc_EOL))	
	{
		if (attributeStack.Top().attrRec.level & slc_DaemonFlag)
			attributeStack.Top().attrRec.daemon->Restore();			

		attributeList[attributeStack.Top().attrType] =
			attributeStack.Top().attrRec;
		attributeStack.Pop();
	}

//	pop off the sentinel...

	attributeStack.Pop();
	level--;

// zero anything that wasn't previously set at this level.

	for (i = 0; i < attributeList.NumItems(); i++)
		if ((attributeList[i].level & slc_LevelMask) > level)
		{
			attributeList[i].data = 0;
			attributeList[i].level = 0;
		}
}

Bool SLContext::IsCurrent(AttrType attrType)
{
	return(attributeList[attrType].level == level);
}

Void SLContext::Clear()
{
	Int i;

	attributeStack.SetSize(0);
	level = 0;

	for (i = 0; i < attributeList.NumItems(); i++)
	{
		attributeList[i].level = 0; 
		attributeList[i].data = 0;
	}	
}

ostream &operator << (ostream &s, const AttrRec &ar)
{
	s << ar.data << "@" << ar.level;
	
	return(s);
}

ostream &operator << (ostream &s, const AttrStackRec &asr)
{
	s << asr.attrRec << ":" << asr.attrType;

	return(s);
}

ostream &operator << (ostream &s, const SLContext &ctx)
{
	s << "SLContext at level " << ctx.level << ". Current attributes: " << endl;
	s << (AttrList &) ctx.attributeList << endl;
	s << "Attribute stack: " << endl << (AttrStackList &) ctx.attributeStack << endl;
	
	return(s);
}


