/*
	File:			SLContext.h

	Function:		State machine for rendering a scene. You can add attributes
					to the context, and also push & pop the current context.
					
	Author(s):		Andrew Willmott

	Copyright:		Copyright (c) 1995-1996, Andrew Willmott
 */

#ifndef __SLContext__
#define __SLContext__


#include "cl/NArray.h"
#include <iostream.h>


#define SC_GET(X) ((sc ## X *) context->Get(a ## X))
// shorthand for fetch & cast: assumes existance of a local variable
// context pointing to the current context.


// --- Attribute types & definition -------------------------------------------

// We don't store the attributes themselves. Instead we store pointers
// to attributes that reside in the scene tree.
// We also define daemons that can be installed to process attributes as they
// are encountered.

typedef Int 			AttrType;
typedef Void 			*AttrPtr;

class AttrDaemon 
{
	public:
	
	virtual Void			Set(AttrPtr attrPtr) = 0;
	virtual AttrPtr			Get() = 0;
	virtual Void			Save() = 0;
	virtual Void 			Restore() = 0;
};

class AttrRec
{
public:
	union
	{
		AttrPtr			data;
		AttrDaemon		*daemon;
	};
	UInt16				level;	// level at which it was last set 
};

class AttrStackRec
{
public:
	AttrRec		attrRec;
	AttrType	attrType;
};

ostream &operator << (ostream &s, const AttrRec &ar);
ostream &operator << (ostream &s, const AttrStackRec &asr);

typedef NArray<AttrRec> AttrList;
typedef NArray<AttrStackRec> AttrStackList;


// --- The SLContext class ----------------------------------------------------


class SLContext
{
public:
	SLContext(Int maxAttributes);

	AttrPtr				Get(AttrType attrType);
						///< Get current attribute of type attrType
	Void				Set(AttrType attrType, AttrPtr attrPtr);
						///< Set attribute of type attrType
	Void 				SetDaemon(AttrType attrType, AttrDaemon *daemon);
	
	Bool				IsCurrent(AttrType attrType);
	
	Void				Push();		///< Push current context
	Void				Pop();		///< Restore last context
	Void				Clear();	///< Clear the context stack

protected:	
	AttrList		attributeList;
	AttrStackList	attributeStack;
	Int				level;			// Current size of the (pseudo) stack

	friend ostream &operator << (ostream &s, const SLContext &ctx);
};


#endif
