/*
	File:		MakeFCH.cc

	Function:	Create Face Cluster Hierarchies

	Author:		Andrew Willmott

	Notes:		
*/

#include "gcl/SceneLang.h"
#include "gcl/Readers.h"
#include "gcl/VecUtil.h"
#include "cl/Heap.h"
#include "cl/Timer.h"
#include "cl/NArray.h"
#include "arg.h"

#include "MeshJoin.h"
#include "Jacobi.h"
#include "Quaternion.h"

#include <sys/resource.h>

#ifdef DEBUG
#define DBG_COUT if (0) cerr
#else
#define DBG_COUT if (0) cerr
#endif

// use floats to save space
#define FCH_FLOAT

// use shape metric
#define FCH_SHAPE

// track normal variation
//#define FCH_NORM

// use original metric. requires FCH_NORM and FCH_SHAPE
//#define FCH_ORIG

// don't run jacobi for flat clusters
//#define FCH_USE_S

// use quaternions to store orientation to save space
#define FCH_USE_QUAT

// --- cluster types ----------------------------------------------------------

#ifdef FCH_FLOAT
typedef Vec3f	FCVec;
typedef Float	FCReal;
#else
typedef Vec3d	FCVec;
typedef Double	FCReal;
#endif

struct SymMat3
{
#ifdef FCH_FLOAT
	Vec3f	diag;
	Vec2f	s1;
	Float	s2;
#else
	Vec3d	diag;
	Vec2d	s1;
	Double	s2;
#endif

	SymMat3 &operator += (const SymMat3 &m)
	{
		diag += m.diag;
		s1 += m.s1;
		s2 += m.s2;
	}
};

Void ConvertMat(const SymMat3 &m, Mat3d &n)
{
	n[0][0] = m.diag[0];	n[0][1] = m.s1[0];		n[0][2] = m.s2;
	n[1][0] = m.s1[0];		n[1][1] = m.diag[1];	n[1][2] = m.s1[1];
	n[2][0] = m.s2;			n[2][1] = m.s1[1];		n[2][2] = m.diag[2];
}

Void ConvertMat(const Mat3d &m, SymMat3 &n)
{
	n.diag[0] = m[0][0];
	n.diag[1] = m[1][1];
	n.diag[2] = m[2][2];
	n.s1[0] = m[0][1];
	n.s1[1] = m[1][2];
	n.s2 = m[0][2];
}


// --- Dual Edge class --------------------------------------------------------


struct DualEdge : public HeapEntry
{
	Int		clus[2];	// ids of cluster at either end
#ifdef FCH_SHAPE
	Float	length;		// length of common boundary
#endif

#ifdef FCH_USE_QUAT
	Quatf	frameQuat;
#else
	SymMat3	frameMat;
#endif
};

typedef NArray<DualEdge> DualEdgeList;

// override MeshJoin class to make list of dual edges

class DEMeshJoin : public MeshJoin
{
public:
	Bool			EdgesMatch(
							FacePtr face1,
							Edge	edge1,
							FacePtr face2,
							Edge	edge2
						);

	DualEdgeList	*dualEdges;
	Face			*faceListBase;
};


Bool DEMeshJoin::EdgesMatch(
							FacePtr face1,
							Edge	edge1,
							FacePtr face2,
							Edge	edge2
						)
{
	Bool	matched = MeshJoin::EdgesMatch(face1, edge1, face2, edge2);

	if (matched)
	{
		DualEdge	newEdge;
		
		newEdge.clus[0] = (face1 - faceListBase);
		newEdge.clus[1] = (face2 - faceListBase);

		dualEdges->Append(newEdge);
	}

	return(matched);
}


// --- Face Cluster class -----------------------------------------------------


struct FaceCluster
{
	Void		InitFromFace(Face &f, PointList &points);
	Void		MergeWith(FaceCluster &c2);
	Float		FindCost(FaceCluster &c2, DualEdge *de);

	// FI info -- keep this as small as possible!
	
	SymMat3		A;			// covariance matrix of points
	FCVec		b;			// average point
	FCReal		c;			// total points

#ifdef FCH_NORM
// this tries to follow garland's code
	SymMat3		D;			// covariance matrix of weighted normals
	FCVec		e;			// sum weighted normal
	FCReal		f;			// sum weights
	FCVec		g;			// sum of normals.
#else
	FCVec		sumAreaNormal;
#endif

#ifdef FCH_SHAPE
	FCReal		length;		// perimeter of the cluster
#endif
	FCReal		area;		// total area of the cluster

	Int			edges;		// first dual edge incident on this face.
	Int			id;			// virtual id of cluster
	DualEdge	*prevDE;	// used to merge edge lists
	
	static Int	sNextID;
};


Int FaceCluster::sNextID = -1;

Void FaceCluster::InitFromFace(Face &face, PointList &points)
{
	Vector 	areaNormal;
	Point	&p0 = points[face.index[0]];
	Point	&p1 = points[face.index[1]];
	Point	&p2 = points[face.index[2]];
	
	CalcTriAreaNormal(p0, p1, p2, areaNormal);
	areaNormal *= 0.5;
	
	area = len(areaNormal);
#ifdef FCH_SHAPE
	length = len(p0 - p1) + len(p2 - p1) + len(p0 -p2);
#endif

	VecTrans	An, Dn;
	Vector		bn, en, gn;
	
	An = oprod(p0, p0);
	An += oprod(p1, p1);
	An += oprod(p2, p2);
	bn = p0;
	bn += p1;
	bn += p2;
	c = 3.0;

	ConvertMat(An, A);
	ConvertVec(bn, b);

#ifdef FCH_NORM
	// garland's code has wi = sqrt(Ai) -- is this right?
	// probably not.
	Dn = oprod(areaNormal, areaNormal) / area / sqrt(area);
	en = areaNormal / sqrt(area);
	f = sqrt(area);
	gn = areaNormal / area;
	
	ConvertMat(Dn, D);
	ConvertVec(en, e);
	ConvertVec(gn, g);
#else
	ConvertVec(areaNormal, sumAreaNormal);
#endif
}

Void FaceCluster::MergeWith(FaceCluster &c2)
{
	area += c2.area;
	A += c2.A;
	b += c2.b;
	c += c2.c;
	// length adjustment done in MergeClusters

#ifdef FCH_NORM
	D += c2.D;
	e += c2.e;
	f += c2.f;
	g += c2.g;
#else
	sumAreaNormal += c2.sumAreaNormal;
#endif
}

Float FaceCluster::FindCost(FaceCluster &c2, DualEdge *de)
{
	Double	result, sumArea, sumPerim, sumProjArea;
	Double	p1, p2, pNew;
	Mat3d	CV, An, An2, Dn, Dn2, evecs;
	Vec3d	evals, bn, bn2, en, en2, gn, gn2;
	Vec3d	n, avgNormal;
	Double	d, eDir, eFit, eShape, cn, fn;

	ConvertMat(A, An);
	ConvertMat(c2.A, An2);
	ConvertVec(b, bn);
	ConvertVec(c2.b, bn2);
	An += An2;
	bn += bn2;
	cn = c + c2.c;

#ifdef FCH_NORM
	ConvertMat(D, Dn);
	ConvertMat(c2.D, Dn2);
	ConvertVec(e, en);
	ConvertVec(c2.e, en2);
	ConvertVec(g, gn);
	ConvertVec(c2.g, gn2);
	Dn += Dn2;
	en += en2;
	gn += gn2;
	fn = f + c2.f;
	avgNormal = norm(gn);
#else
	ConvertVec(sumAreaNormal + c2.sumAreaNormal, avgNormal);
	sumProjArea = len(avgNormal);
	avgNormal /= sumProjArea;
#endif

	sumArea = area + c2.area;
	
#ifdef FCH_USE_S
	if (sumProjArea > sumArea * 0.5)
	{
		n = avgNormal;
		evecs = vl_1;
	}
	else
	{
#endif
		CV = An - oprod(bn, bn) / cn;
		Jacobi(CV, evals, evecs);

		if (dot(evecs[2], avgNormal) < 0)
			evecs[2] = -evecs[2];
			
		// ensure frame is right-handed
		if (det(evecs) < 0)
			evecs[1] = -evecs[1];

		if (sumProjArea > sumArea * 0.5)
			n = avgNormal;
		else
			n = evecs[2];

#ifdef FCH_USE_S
	}
#endif
	
	d = -dot(n, bn) / cn;

#ifdef FCH_USE_QUAT
	de->frameQuat = MakeQuatf(evecs);
//	de->frameQuat = MakeQuat(evecs);
#else
	ConvertMat(evecs, de->frameMat);
#endif

	eFit = dot(n, An * n) / cn - sqr(d);

#ifdef FCH_NORM
	eDir = dot(avgNormal, Dn * avgNormal) 
			- 2.0 * dot(gn, avgNormal) + fn;
	eDir /= sumArea;
#endif

#ifdef FCH_SHAPE
	sumPerim = length + c2.length - 2.0 * de->length;
#else
	sumPerim = sqrt(sumArea);
#endif
	
#ifdef FCH_ORIG
	p1 = sqr(length) / area;
	p2 = sqr(c2.length) / c2.area;


	pNew = sqr(sumPerim) / sumArea;
    eShape = 1.0 - 0.5 * Max(p1, p2) / pNew;

	result = eFit + eDir + eShape;
#else
	result = (eFit / sumArea + 0.01) * (sumArea / sumProjArea) * sqr(sumPerim);
#endif
		
	Assert(vl_is_finite(result), "infinite result!");

	return(-result);
}


// --- Clusterer class --------------------------------------------------------


typedef NArray<Face> FaceList;
typedef NArray<FaceCluster>	ClusterList;

struct Clusterer : public Decimator
{		
	Void			ClusterScene(scScenePtr scene);

	Void			MergeClusters(Int fc1, Int fc2, DualEdge *mergeDE);
	Void			MergeEdgeLists(Int fc1, Int fc2, DualEdge *mergeDE);

	Void			ExtractDualEdges(scScenePtr scene);
	Void			WriteBaseModel();
	Void 			CreateLeafClusters();
	Void			InitDualEdge(DualEdge *de);
	
	Float			FindMemUse();
	Void			HandlePoly(Int numVertices, Int vertices[], Int changed);

	FaceList		faces;			// list of initial faces
	Int				numFaces;
	DEMeshJoin		*joiner;

	ClusterList		faceClusters;	// the face clusters
	DualEdgeList	dualEdges;		// dual edges
	IndexList		edgeListNext;	// cluster-to-dualEdge mapping
	Heap			dualHeap;		// sort-by-cost heap

	ProgramTimer	timer;
};


/*
	NOTE

	the edgeListNext list stores next fields for edge list from faces.
	it is twice as large as the list of dual edges.

	"links" in the bottom part of the array belong to faces pointed at
	by the clus[0] side of the edge.

	"links" in the top part of the array belong to faces pointed at by
	the clus[1] side of the edge.

	thus, the edges field of the cluster both points to the first link
	in the list of dual edges belonging to that face, and encodes which
	cluster field of the edge it corresponds to.

*/

Void Clusterer::InitDualEdge(DualEdge *de)
// just assign initial border length
{
#ifdef FCH_SHAPE
	Face 	&f0 = faces[de->clus[0]];
	Face 	&f1 = faces[de->clus[1]];
	Int		i;
	
	for (i = 0; i < 3; i++)
		if (f0.adjFace[i] == &f1)
		{
			Point	&p0 = (*pointsAcc)[f0.index[i]];
			Point	&p1 = (*pointsAcc)[f0.index[(i + 1) % 3]];

			de->length = len(p0 - p1);
			break;
		}

	DBG_COUT << "border = " << de->length << endl;
#endif
}

Void Clusterer::CreateLeafClusters()
{
	Int		i;
	Int		*head;
	
	faceClusters.SetSize(faces.NumItems());

	for (i = 0; i < faceClusters.NumItems(); i++)
	{
		faceClusters[i].InitFromFace(faces[i], *pointsAcc);
		faceClusters[i].id = i;
		faceClusters[i].edges = -1;
	}
	
	// now, create edge lists for face clusters

	edgeListNext.SetSize(dualEdges.NumItems() * 2);
	
	for (i = 0; i < dualEdges.NumItems(); i++)
	{
		// add to the edge lists of clusters at either end

		head = &faceClusters[dualEdges[i].clus[0]].edges;
		// find place to insert, ordered by other cluster id
		while (*head >= 0 && 
			dualEdges[*head % dualEdges.NumItems()]
				.clus[(*head / dualEdges.NumItems()) ^ 1] > dualEdges[i].clus[1])
			head = &edgeListNext[*head];

		edgeListNext[i] = *head;
		*head = i;			

		head = &faceClusters[dualEdges[i].clus[1]].edges;
		// find place to insert, ordered by other cluster id
		while (*head >= 0 && 
			dualEdges[*head % dualEdges.NumItems()]
				.clus[(*head / dualEdges.NumItems()) ^ 1] > dualEdges[i].clus[0])
			head = &edgeListNext[*head];

		edgeListNext[i + dualEdges.NumItems()] = *head;
		*head = i + dualEdges.NumItems();			

		// set up common boundary length field
		InitDualEdge(&dualEdges[i]);
	}

	FaceCluster::sNextID = faces.NumItems();

#ifdef DEBUG
	DBG_COUT << "LIST CHECK" << endl;

	for (i = 0; i < faceClusters.NumItems(); i++)
	{
		Int		ed;
		
		DBG_COUT << "edges for " << i << "..." << endl;
		for (ed = faceClusters[i].edges; ed >= 0; ed = edgeListNext[ed])
		{
			DualEdge *de = &dualEdges[ed % dualEdges.NumItems()];
			DBG_COUT << "link " << de->clus[0] << " <> " << de->clus[1] << endl;
		}
		DBG_COUT << "area = " << faceClusters[i].area << endl;
	}
#endif
}

Void Clusterer::MergeClusters(Int fc1, Int fc2, DualEdge *mergeDE)
{
	FaceCluster		&cluster1 = faceClusters[fc1];
	FaceCluster		&cluster2 = faceClusters[fc2];
	Int			 	i;

	DBG_COUT << "merging " << fc1 << " and " << fc2 << endl;
	DBG_COUT << "----------------------" << endl;

	// merge cluster data
	cluster1.MergeWith(cluster2);
#ifdef FCH_SHAPE
	// update perimeter
	cluster1.length += cluster2.length - 2.0 * mergeDE->length;
#endif
	// merge edge lists
	MergeEdgeLists(fc1, fc2, mergeDE);	
}

static Int		gMaxEdges = 0;
static Int		gTotEdges = 0;
static Int		gTotSamples = 0;

Void Clusterer::MergeEdgeLists(Int fc1, Int fc2, DualEdge *mergeDE)
{
	FaceCluster		&cluster1 = faceClusters[fc1];
	FaceCluster		&cluster2 = faceClusters[fc2];
	Int			 	i;
	Int				el, end, otherFC;
	DualEdge		*de;
	Int				*tail;
	Int				numEdges = 0;		
	
	// mark clusters adjacent to cluster1 with pointers to the
	// corresponding dual edge
	for (el = cluster2.edges; el >= 0; el = edgeListNext[el])
	{
		de = &dualEdges[el % dualEdges.NumItems()];
		end = el / dualEdges.NumItems();
		otherFC = de->clus[end ^ 1];
		faceClusters[otherFC].prevDE = 0;
	}

	// traverse cluster1 links, updating costs as we go
	tail = &cluster1.edges;

	for (el = *tail; el >= 0; el = edgeListNext[el])
	{
		de = &dualEdges[el % dualEdges.NumItems()];
		end = el / dualEdges.NumItems();
		
		DBG_COUT << "link0 " << de->clus[0] << " <> " << de->clus[1] << endl;

		if (de->heapIdx < 0)
		{
			DBG_COUT << "  UNUSED" << endl;
			
			// edge no longer in use : delete!
			*tail = edgeListNext[el];
			continue;
		}

		otherFC = de->clus[end ^ 1];
		faceClusters[otherFC].prevDE = de;
		
		de->cost = cluster1.FindCost(faceClusters[otherFC], de);
		dualHeap.Update(de);

		tail = &edgeListNext[el];

		numEdges++;
	}

	// append cluster2 edge list
	*tail = cluster2.edges;

	// and traverse it, merging duplicate links and updating costs
	for (el = *tail; el >= 0; el = edgeListNext[el])
	{
		de = &dualEdges[el % dualEdges.NumItems()];
		end = el / dualEdges.NumItems();

		DBG_COUT << "link1 " << de->clus[0] << " <> " << de->clus[1] << endl;
		
		if (de->heapIdx < 0)
		{
			// edge no longer in use
			DBG_COUT << "UNUSED" << endl;

			*tail = edgeListNext[el];
			continue;
		}

		otherFC = de->clus[end ^ 1];

		if (faceClusters[otherFC].prevDE)
		{
			DBG_COUT << "REJECTED" << endl;

			// accumulate perimeter
#ifdef FCH_SHAPE
			faceClusters[otherFC].prevDE->length += de->length;
#endif
			dualHeap.Delete(de);
			*tail = edgeListNext[el];
			continue;
		}
							
		// rewrite edge to point to new cluster
		de->clus[end] = fc1;

		// update cost
		de->cost = cluster1.FindCost(faceClusters[otherFC], de);
		dualHeap.Update(de);

		tail = &edgeListNext[el];

		numEdges++;
	}

	*tail = -1;

	gTotEdges += numEdges;
	gTotSamples++;
	gMaxEdges = Max(numEdges, gMaxEdges);

#ifdef DEBUG
	DBG_COUT << "List is now:" << endl;
	for (el = cluster1.edges; el >= 0; el = edgeListNext[el])
	{
		de = &dualEdges[el % dualEdges.NumItems()];
		end = el / dualEdges.NumItems();

		DBG_COUT << "  link " << de->clus[0] << " <> " << de->clus[1] << endl;
	}
#endif

	cluster1.id = FaceCluster::sNextID;
	cluster2.id = -1;		// now unused

	DBG_COUT << " -> " << cluster1.id << endl;
	DBG_COUT << endl;
	
	FaceCluster::sNextID++;
}

Void Clusterer::ExtractDualEdges(scScenePtr scene)
{
	FindDecInfo		decInfo;
	Int				i, numSurfPoints;
	
	scene->Decimate(decInfo, DecTris);
	numFaces = 0;
	faces.SetSize(decInfo.numTris);
	scene->Decimate(SELF, DecTris | DecUseMaster);
	delete scene;

	// Create dual edge list

	joiner = new DEMeshJoin();
	joiner->faceListBase = faces.Ref();
	joiner->dualEdges = &dualEdges;

	joiner->Init(pointsAcc->NumItems());
	DBG_COUT << "group start" << endl;
	joiner->StartGroup();
	DBG_COUT << "adding tris" << endl;
	for (i = 0; i < faces.NumItems(); i++)
		joiner->AddTriangle(&faces[i]);
	DBG_COUT << "group end" << endl;
	joiner->EndGroup();

	numSurfPoints = joiner->Finished();

	cerr << "Dual edges: " << dualEdges.NumItems() << endl;
	cerr << "Surface points: " << numSurfPoints << endl;

	cout << "# dual_edges " << dualEdges.NumItems() << endl;

	delete joiner;
}

Void Clusterer::WriteBaseModel()
{
	Int		i;
	
	cerr << "points: " << pointsAcc->NumItems() << endl;
	cerr << "faces: " << faces.NumItems() << endl;

	cout << "# points " << pointsAcc->NumItems() << endl;
	cout << "# faces " << faces.NumItems() << endl;
	
	for (i = 0; i < pointsAcc->NumItems(); i++)
		printf("v %g %g %g\n", 
			(*pointsAcc)[i][0], (*pointsAcc)[i][1], (*pointsAcc)[i][2]);
	for (i = 0; i < faces.NumItems(); i++)
		printf("f %d %d %d\n", 
			faces[i].index[0] + 1, faces[i].index[1] + 1, faces[i].index[2] + 1);
			  	
	cerr << "wrote base model." << endl;
}

Void Clusterer::ClusterScene(scScenePtr scene)
{
	Int				i, clusterTrees, lastLevel;
	Float			time1, time2, time3, time4, mem1, mem2;
#ifdef FCH_USE_QUAT
	Mat3f			frame;
#else
	Mat3d			frame;
#endif
	
	timer.StartTimer();
	
	// create the dual edges & faces
	ExtractDualEdges(scene);
	time1 = timer.GetTimer();

	WriteBaseModel();
	time2 = timer.GetTimer();
		
	mem1 = FindMemUse();

	// initialise the leaf face clusters!
	CreateLeafClusters();

#ifdef FCH_OUTPUT_LEAVES
	timer.StopTimer();

	for (i = 0; i < faces.NumItems(); i++)
	{
		Vector 	areaNormal, va;
		Point	&p0 = (*pointsAcc)[faces[i].index[0]];
		Point	&p1 = (*pointsAcc)[faces[i].index[1]];
		Point	&p2 = (*pointsAcc)[faces[i].index[2]];
		
		CalcTriAreaNormal(p0, p1, p2, areaNormal);

		cout << "f^ 0 0" << endl;

		cout << "fe ";
		va = norm(areaNormal);
		cout << va[0] << ' ' << va[1] << ' ' << va[2] << ' ';
		va = norm(p1 - p0);
		cout << va[0] << ' ' << va[1] << ' ' << va[2] << ' ';
		va = norm(cross(areaNormal, va));
		cout << va[0] << ' ' << va[1] << ' ' << va[2] << ' ';
		cout << endl;
	}
	timer.ContTimer();
#endif

	// don't need these any more
	delete pointsAcc;
	pointsAcc = 0;
	faces.Clear();

	cerr << "creating heap" << endl;
	
	// create the heap!
	for (i = 0; i < dualEdges.NumItems(); i++)
	{
		DBG_COUT << "dual edge " << dualEdges[i].clus[0] 
			 << " <-> " << dualEdges[i].clus[1] << endl;

		dualEdges[i].cost =
			faceClusters[dualEdges[i].clus[0]].
				FindCost(faceClusters[dualEdges[i].clus[1]], &dualEdges[i]);

		dualHeap.Insert(&dualEdges[i]);
	}
	time3 = timer.GetTimer();

	// iterate!

	clusterTrees = 0;
	
	cerr << "processing " << dualHeap.NumItems() << " edges..." << endl;
	lastLevel = (dualHeap.NumItems() / 20000) * 20000;
	
	while (dualHeap.NumItems() > 0)
	{
		DualEdge *de = (DualEdge*) dualHeap.RemoveMax();

		if (dualHeap.NumItems() < lastLevel)
		{
			cerr << "dual edges left: " << dualHeap.NumItems() << endl;
			lastLevel -= 20000;
		}
		
		Assert(faceClusters[de->clus[0]].id >= 0, "bad left cluster");
		Assert(faceClusters[de->clus[1]].id >= 0, "bad right cluster");

		if (de->clus[0] == de->clus[1])
		{
			clusterTrees++;
			continue;
		}

#ifndef FCH_NO_OUT
		cout << "f^ " << faceClusters[de->clus[0]].id + 1
			 << " " << faceClusters[de->clus[1]].id + 1 << endl;
		
		cout << "fc " << de->cost << endl;

#ifdef FCH_USE_QUAT
		frame.MakeRot(de->frameQuat);
#else
		ConvertMat(de->frameMat, frame);
#endif

		cout << "    fe " 
			 << frame[0][0] << ' ' << frame[0][1] << ' ' << frame[0][2] << ' '
			 << frame[1][0] << ' ' << frame[1][1] << ' ' << frame[1][2] << ' '
			 << frame[2][0] << ' ' << frame[2][1] << ' ' << frame[2][2] << ' '
			 << endl;

		DBG_COUT << "cost = " << de->cost << endl;
#endif

		MergeClusters(de->clus[0], de->clus[1], de);

#ifdef DEBUG
		for (i = 0; i < dualEdges.NumItems(); i++)
		{
			CheckRange(dualEdges[i].clus[0], 0, faceClusters.NumItems(), "bad clus");
			CheckRange(dualEdges[i].clus[1], 0, faceClusters.NumItems(), "bad clus");
		}
#endif
	}
	
	time4 = timer.GetTimer();
	mem2 = FindMemUse();
	
	cerr << "Total cluster trees: " << clusterTrees << endl;
	cerr << "Create dual edges: " << time1 << endl;
	cerr << "Write base model: " << time2 - time1 << endl;
	cerr << "Init clusters & heap: " << time3 - time2 << endl;
	cerr << "Clustering: " << time4 - time3 << endl;
	cerr << "Total Time Taken: " << time4 << endl;
	cerr << "Prep memory use (MB): " << mem1 << endl;
	cerr << "Cluster memory use (MB): " << mem2 << endl;

	cout << "# find_dual_edges " << time1 << endl;
	cout << "# init " << time3 - time2 << endl;
	cout << "# clustering " << time4 - time3 << endl;
	cout << "# total " << time4 << endl;
	cout << "# join_memory " << mem1 << endl;
	cout << "# memory " << mem2 << endl;

	cerr << "Max dual edges per cluster = " << gMaxEdges << endl;
	cerr << "Average dual edges per cluster = " << Double(gTotEdges) / gTotSamples << endl;

	cout << "# max_de " << gMaxEdges << endl;
	cout << "# avg_de " << Double(gTotEdges) / gTotSamples << endl;

	
}

Float Clusterer::FindMemUse()
{
	Int		mem;
	
	mem = sizeof(DualEdge) * dualEdges.NumItems();
	mem += sizeof(FaceCluster) * faceClusters.NumItems();
	mem += sizeof(Void*) * dualHeap.NumItems();
	mem += sizeof(Index) * edgeListNext.NumItems();
	mem += sizeof(Face) * faces.NumItems();
	if (pointsAcc)
		mem += pointsAcc->NumItems() * sizeof(Point);
	mem += sizeof(Clusterer);
	
	return(mem / sqr(1024.0));
}

Void Clusterer::HandlePoly(Int numVertices, Int vertices[], Int changed)
{
	if (numVertices < 0)
		return;

	faces[numFaces].index[0] = vertices[0];
	faces[numFaces].index[1] = vertices[1];
	faces[numFaces].index[2] = vertices[2];

	numFaces++;
}

static Void PrintSceneInfo(scScenePtr scene)
{
	FindDecInfo	decInfo;
	
	scene->Decimate(decInfo, DecTris);
	
	cerr << "polygons " << decInfo.numPolys << endl;
	cerr << "triangles " << decInfo.numTris << endl;
}

int main(int argc, char *argv[])
{
	StrConst		scenePath;
	FileName		sceneFile;
	Arg_form		*arg_format;
	scScenePtr		scene;
	Clusterer		clusterer;
		
	// set options
	
	arg_format = arg_to_form(0,
		"", 										"usage: make-fch [options]",
		
		"%S",	&scenePath,							"Specify scene file",
			
		0);

	// No arguments: print usage
	if (argc == 1)
	{
		arg_form_print(arg_format);
		exit(1);
	}
	// parse arguments
	if (arg_parse_argv(argc, argv, arg_format) < 0)
		exit(1);

	sceneFile.SetPath(scenePath);
	cerr << "Processing " << scenePath << endl;	

	slInit();

	scene = SceneReader::Load(sceneFile);
	if (!scene)
		return(1);

	PrintSceneInfo(scene);
	clusterer.ClusterScene(scene);

	return(0);
}
